package raytracer;

import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.File;

import javax.swing.*;
import javax.imageio.ImageIO;

import raytracer.internal.Color;
import raytracer.internal.Constants;

class Renderer extends JPanel implements Runnable {
    public static final int NO_SCENE_DEFINED = 10233;
    public static final int SUCCEEDED = 22103;
    public static final int CANCELLED = 32413;
    public static final int PAUSED = 66234;
    public static final int NOT_PAUSED = 82254;

    private int width;
    private int height;
    private int recursionDepth;
    private BufferedImage image;

    private Scene scene;
    private boolean allowRender = false;
    private ControlFrame observer = null;
    private boolean cancelled = false;
    private boolean rendering = false;
    private boolean paused = false;
    private int pausedY;

    public Renderer() {
        init(null, Main.DEPTH, Main.WIDTH, Main.HEIGHT);
    }

    public Renderer(String scenename, int width, int height, int recursionDepth) {
        init(scenename, recursionDepth, width, height);
    }

    synchronized public void setSize(int width, int height) {
        waitForCancelRendering();
        this.width = width;
        this.height = height;
        if ((image == null) || (image.getWidth() != width) || (image.getHeight() != height))
            image = new BufferedImage(width, height, BufferedImage.TYPE_3BYTE_BGR);
        setPreferredSize(new Dimension(width, height+10));
    }

    private void waitForCancelRendering() {
        if (rendering) {
            cancelRendering();
            while (rendering)
                try {
                    wait();
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
        }
    }

    synchronized public void setRecursionDepth(int recursionDepth) {
        waitForCancelRendering();
        if (recursionDepth < 0)
            this.recursionDepth = Constants.MAX_RECURSION_DEPTH;
        else
            this.recursionDepth = recursionDepth;
    }

    synchronized public void setSceneName(String sceneName) {
        waitForCancelRendering();
        if (sceneName == null)
            this.scene = null;
        else
            this.scene = SceneReader.parseFile(sceneName);
    }

    private void init(String sceneName, int recursionDepth, int width, int height) {
        setSceneName(sceneName);
        setRecursionDepth(recursionDepth);
        setSize(width, height);

        // start the thread
        new Thread(this).start();
    }

    public void run() {
        allowRender = false;
        while (true) {
            while (!allowRender)
                synchronized (this) {
                    try {
                        wait();
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }
            renderPicture();
            repaint();
        }
    }

    public void paint(java.awt.Graphics g) {
        super.paint(g);
        if ((g != null) && (image != null))
            g.drawImage(image, 0, 0, null);
    }

    synchronized void startRendering(ControlFrame caller) {
        ControlFrame.addConsoleText("Rendering started", Constants.commentColor);
        observer = caller;
        allowRender = true;
        notifyAll();
    }

    synchronized void cancelRendering() {
        ControlFrame.addConsoleText("Rendering cancelled", Constants.commentColor);
        cancelled = true;
    }

    synchronized void resumeRendering(ControlFrame caller) {
        if (!paused) {
            ControlFrame.addConsoleText("Rendering not paused", Constants.failColor);
            if (caller != null)
                caller.renderFinished(NOT_PAUSED);
            return;
        }
        ControlFrame.addConsoleText("Rendering resumed", Constants.commentColor);
        allowRender = true;
        observer = caller;
        notifyAll();
    }

    synchronized void pauseRendering() {
        if (rendering) {
            ControlFrame.addConsoleText("Rendering paused", Constants.commentColor);
            paused = true;
        }
    }

    void renderPicture() {
        allowRender = false;
        cancelled = false;
        if (scene == null) {
            ControlFrame.addConsoleText("No scene defined", Constants.failColor);
            repaint();
            if (observer != null)
                observer.renderFinished(NO_SCENE_DEFINED);
            return;
        }
        if (!paused) {
            pausedY = 0;
        }
        paused = false;
        rendering = true;

        float dx = scene.getWindowWidth()/(float)(width - 1);
        float dy = scene.getWindowHeight()/(float)(height - 1);
        final float w2 = ((float) width )/2f;
        final float h2 = ((float) height)/2f;
        Color col;
        for (int y = pausedY; (y < height) && (!cancelled) && (!paused); y++) {
            pausedY = y;
            for (int x = 0; x < width; x++) {
                col = scene.renderColor(dx*(w2-x), dy*(h2-y), recursionDepth);
                col.normalize();
                image.setRGB(x, y, col.getRGBAsInt());
            }
            if (observer != null)
                observer.setPercent(y*100/height);
            if (y % 10 == 0)
                repaint();
        }

        repaint();

        if (observer != null) {
            if (cancelled) {
                observer.renderFinished(CANCELLED);
            } else if (paused) {
                observer.renderFinished(PAUSED);
            } else {
                ControlFrame.addConsoleText("Rendering done", Constants.commentColor);
                observer.renderFinished(SUCCEEDED);
            }
            observer = null;
        }

        rendering = false;
        synchronized (this) {
            notifyAll();
        }
        //@todo remove debug print
        //scene.print();
    }

    public void saveImage(File f) {
        if (f == null)
            return;
        try {
            if (!ImageIO.write(image, "JPEG", f))
                JOptionPane.showMessageDialog(this, "Cannot save file !\n"+f.getAbsolutePath());
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
}
