package raytracer;

import java.util.Iterator;

import raytracer.internal.Color;
import raytracer.internal.Constants;
import raytracer.internal.Hit;
import raytracer.internal.Point;
import raytracer.internal.Ray;
import raytracer.internal.Vector;
import raytracer.objects.*;

/**
 * 
 * A scene contains the surfaces which should be considered by the renderer
 *
 */
public class Scene {
    private Color ambientColor = new Color(.2f, .2f, .2f);
    private Color backgroundColor = new Color(0, 0, 0);
    private Point camPosition = new Point(0, 0, -5);
    private Point lookAt = new Point(0, 0, 0);
    private Vector lookUp = new Vector(0, 1, 0);
    private Vector camDir;
    private Vector lookRight;
    private float refractionIndex = 1;
    private float windowWidth = 1;
    private float windowHeight = 1;
    private java.util.Vector surfaces;
    private java.util.Vector lights;
    //necessary for the loader
    private final java.util.HashMap store = new java.util.HashMap();

    public final Color white = new Color(1, 1, 1);

    public Scene() {
        surfaces = new java.util.Vector();
        lights = new java.util.Vector();
    }

    /**
     * Is called by the SceneReader after everything is read. Can be used to prepare some calculations, create octrees...
     */
    public void prepare() {
        lookUp = lookUp.normalize();
        camDir = lookAt.sub(camPosition).normalize();
        lookRight = lookUp.crossProduct(camDir).normalize().invert();

        print();
    }

    void print() {
        // print out surfaces
        for (int i = 0; i < surfaces.size(); i++) {
            Surface surface = (Surface) surfaces.elementAt(i);
            if (surface instanceof Sphere)
                System.out.println("sphere r="+((Sphere)surface).getRadius());
            else
                System.out.println("obj name="+surface.getName());
        }
        // print out lights
        for (int i = 0; i < lights.size(); i++) {
            Light light = ((Light) lights.elementAt(i));
            System.out.println("light name="+light.getName());
        }

        for (Iterator it = store.values().iterator(); it.hasNext();) {
            NamedObject namedObject = (NamedObject) it.next();
            if (namedObject instanceof Quadric)
                ((Quadric)namedObject).tst();
        }
    }


    public float getWindowWidth() {
        return windowWidth;
    }

    public float getWindowHeight() {
        return windowHeight;
    }

    //methods for the dynamic loading interface
    public void setRefractionIndex(float refractionIndex) {
        this.refractionIndex = refractionIndex;
    }

    public void setAmbient(float r, float g, float b) {
        ambientColor = new Color(r, g, b);
    }

    public void setBackground(float r, float g, float b) {
        backgroundColor = new Color(r, g, b);
    }
    public void setObject(Surface s) {
        if ((s instanceof Group) || (s instanceof GroupCopy)) {
            final Surface[] objects = ((Group) s).getPrimitiveSurfaces();
            if (objects != null)
                for (int i = 0; i < objects.length; i++) {
                    surfaces.add(objects[i]);
                    System.out.println(""+i+" "+objects[i].getName());
                }
            //surfaces.add(s);
        } else
            surfaces.add(s);
    }
    public void setLight(Light l) {
        lights.add(l);
    }
    public void setCamerapos(float x, float y, float z) {
        camPosition = new Point(x, y, z);
    }
    public void setLookat(float x, float y, float z) {
        lookAt = new Point(x, y, z);
    }
    public void setLookup(float x, float y, float z) {
        lookUp = new Vector(x, y, z);
    }
    public void setWindowsize(float x, float y) {
        windowWidth = x;
        windowHeight = y;
    }

    public Color renderColor(float x, float y, float recursionDepth) {
        final Point point = lookAt.add(lookRight.scale(x)).add(lookUp.scale(y));
        final Vector direction = point.sub(camPosition);
        final Ray ray = new Ray(camPosition, direction);

        return trace(ray, 1f, recursionDepth);
    }

    private Hit sceneRayIntersection(final Ray ray) {
        Hit closestHit = Hit.NO_INTERSECTION;

        for (int i = 0; i < surfaces.size(); i++) {
            Surface surface = (Surface) surfaces.elementAt(i);
            final Hit currentHit = surface.intersectRay(ray);
            if (currentHit.getT() < closestHit.getT())
                closestHit = currentHit;
        }
        return closestHit;
    }

    /**
     * Intersect scene with ray.
     * @param ray Incident ray.
     * @param weight The weight of the computed color.
     * @param recursionDepth
     * @return the backgroundColor if no surface hit or shaded surface if so.
     */
    private Color trace(final Ray ray, float weight, float recursionDepth) {
        final Hit hit = sceneRayIntersection(ray);
        if (hit.getSurface() != null)
            return shade(hit, weight, recursionDepth);
        return new Color(backgroundColor);
    }

    /**
     * Compute Phong shading color after ray hit an object
     * @param hit Scene-ray intersection result.
     * @param weight The weight of the computed color.
     * @param recursionDepth
     * @return Computed Phong shaded color.
     */
    private Color shade(final Hit hit, final float weight, final float recursionDepth) {
        Vector normal = hit.getNormal();
        if (normal.dotProduct(hit.getRay().direction) > 0)
            normal.scale(-1);

        Color diffuseAndSpecularColor = new Color(0, 0, 0);
        Color reflexionColor = new Color(0, 0, 0);
        Color transmissionColor = new Color(0, 0, 0);
        final Material surfaceMaterial = hit.getMaterial();

        // compute diffuse and specular color
        for (int i = 0; i < lights.size(); i++) {
            Light light = (Light) lights.elementAt(i);

            // from point to the light
            Vector lightDir = light.getPosition().sub(hit.getPoint());
            float lightDistance = lightDir.length();
            lightDir = lightDir.normalize();
            final Vector specDir = lightDir.reflect(normal);
            // from point to the camera
            final Vector eyeDir = camPosition.sub(hit.getPoint()).normalize();

            final float ks = (float) Math.pow(specDir.dotProduct(eyeDir), surfaceMaterial.getSpecularExponent());
            final float kd = normal.dotProduct(lightDir);
            if (!shadow(new Ray(hit.getPoint(), lightDir), lightDistance)) {
                // add specular
                if (ks > 0)
                    diffuseAndSpecularColor.add(surfaceMaterial.getSpecularColor(), light.getColor(), ks);
                // add diffuse
                if (kd > 0)
                    diffuseAndSpecularColor.add(surfaceMaterial.getDiffuseColor(), light.getColor(), kd);
                //diffuseAndSpecularColor.mul(white, 1f / lightDistance * lightDistance);
            }
        }

        // compute reflexion and transmission
        if (recursionDepth > 0) {
            // compute specular reflexion ray color
            final float wr = surfaceMaterial.getSpecularColor().influence() * weight;
            if (wr > Constants.LIGHT_INFLUENCE_EPSILON) {
                // reflected ray
                final Ray newRay = new Ray(
                        hit.getPoint(),
                        hit.getRay().direction.reflect(normal));
                reflexionColor = trace(newRay, recursionDepth - 1, wr);
            }

            // compute transmission
            final float wt = surfaceMaterial.getRefractionIndex() * weight;
            if (wt > Constants.LIGHT_INFLUENCE_EPSILON) {
                final Vector dir = transmissionDir(hit);
                if (dir == null) {
                    transmissionColor = new Color(0, 0, 0);
                } else {
                    // refracted ray
                    final Ray newRay = new Ray(hit.getPoint(), dir);
                    transmissionColor = trace(newRay, recursionDepth - 1, wt);
                }
            }

        }

        // init color to the ambient color
        Color color = new Color(ambientColor);

        // diffuse and specular color
        color.add(diffuseAndSpecularColor);

        // reflexion color
        // @todo change to diffuse/specular reflexion ???
        color.add(surfaceMaterial.getReflectionColor(), reflexionColor);

        // transmission color
        color.add(surfaceMaterial.getTransmissionColor(), transmissionColor);

        // texture color
        color.mul(hit.getTextureColor());

        return color;
    }

    /**
     * @param ray
     * @param length
     * @return true if point in shadow (ray intersects an object before length)
     */
    private boolean shadow(Ray ray, float length) {
        Hit hit = sceneRayIntersection(ray);

        if (hit == null)
            return false;

        if (hit.getT() > length - Constants.OUTSIDE_EPSILON)
            return false;

        return true;
    }

    /**
     * @param hit
     * @return transmission direction (refraction direction)
     */
    private Vector transmissionDir(Hit hit) {
        float n1, n2;
        if (hit.entering()) { // entering object
            n1 = refractionIndex;
            n2 = hit.getMaterial().getRefractionIndex();
        } else { // exiting object
            n2 = refractionIndex;
            n1 = hit.getMaterial().getRefractionIndex();
        }

        final Vector dir = hit.getRay().direction;
        float eta = n1 / n2;
        float cosTheta = 0f - dir.dotProduct(hit.getNormal());
        float cosTheta2 = 1f - eta*eta*(1f - cosTheta*cosTheta);
        if (cosTheta2 < 0) { // total internal reflexion
            return null;
        }
        return dir.scale(eta).scaleAdd(
                ((float) (eta * cosTheta - Math.sqrt(cosTheta2))),
                hit.getNormal());
    }

    public void clearStore() {
        store.clear();
    }
    public void store(String name, NamedObject obj) {
        store.put(name, obj);
    }
    public boolean isObjectName(String word) {
        return store.containsKey(word);
    }
    public Object getObjectByName(String name) {
        return store.get(name);
    }
}