package raytracer;

import java.io.FileReader;
import java.io.IOException;
import java.io.StreamTokenizer;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import raytracer.objects.NamedObject;
import raytracer.internal.Color;
import raytracer.internal.Constants;

class SceneReader {
    static Scene parseFile(String filename) {
		try {
			final Scene scene = new Scene();
			scene.clearStore();
			final StreamTokenizer tok = new StreamTokenizer(new FileReader(filename));
			tok.eolIsSignificant(false);
			tok.parseNumbers();
			tok.slashSlashComments(true);
			tok.slashStarComments(true);

			boolean textIsNewSurface = true;
			Class currentClass = null;
			Object currentInstance = null;
			java.util.Vector argumentVector = new java.util.Vector();
			String methodName = null;
			boolean ommitCurrentObject = false;

			do {
				switch (tok.nextToken()) {

					case StreamTokenizer.TT_WORD :
						if (textIsNewSurface) { //"first" words are always new objects or "Scene"
							if (tok.sval.equals("Scene")) { //special case "Scene"
								currentClass = Scene.class;
								currentInstance = scene;
							}
							else { //otherwise: dynamically load classes 
								try {
									currentClass = Class.forName("raytracer.objects." + tok.sval);
									currentInstance = currentClass.newInstance();
									tok.nextToken(); //there must be a second one with the name!
									 ((NamedObject)currentInstance).setName(tok.sval);
								}
								catch (ClassNotFoundException ex) {
                                    writeError("Object type \"" + tok.sval + "\" not found.");
 									ommitCurrentObject = true;
								}
								catch (InstantiationException ex) {
                                    writeError("Unable to create surface of type \""
                                                                                + currentClass.getName()
                                                                                + "\".");
									ommitCurrentObject = true;
								}
							}
							textIsNewSurface = false;
						}
						else if (
							!ommitCurrentObject
								&& (scene.isObjectName(tok.sval) // should be an object-parameter
								|| tok.sval.endsWith("-"))) {    // or a string-parameter
							Object[] objectParaArray = { scene.getObjectByName(tok.sval)};
							if (tok.sval.endsWith("-")) {
								methodName = getMethodName(tok.sval);
								tok.nextToken();
								try {
									objectParaArray[0] = new String(tok.sval);
								} catch (Exception e) {
									objectParaArray[0] = new String();
								}
							}
							try {
								final Method[] methods = currentClass.getMethods();
								boolean success = false;
								for (int i = 0; i < methods.length; i++) {
									final Method method = methods[i];
									final Class[] paraTypes = method.getParameterTypes();
									if (paraTypes.length == 1
										&& methodName.equals(method.getName())
										&& paraTypes[0].isAssignableFrom(
											objectParaArray[0].getClass())) {
										method.invoke(currentInstance, objectParaArray);
										success = true;
										break;
									}
								}
								if (!success)
									writeError("Object type \""
											+ currentClass.getName()
											+ "\" does not support method \""
											+ methodName
											+ "\" with parameter of type \""
											+ objectParaArray[0].getClass().getName()
											+ "\".");
							}
							catch (InvocationTargetException ex) {
								writeError("Unable to invoke method: "+methodName+" " + ex);
							}
						}
						else { //other words are method calls or "end"
							if (!ommitCurrentObject && !argumentVector.isEmpty()) {
								Class[] paraClassArray = new Class[argumentVector.size()];
								for (int i = 0; i < paraClassArray.length; i++)
									paraClassArray[i] = Float.TYPE;
								try {
									currentClass.getMethod(methodName, paraClassArray).invoke(
										currentInstance,
										argumentVector.toArray());
								}
								catch (NoSuchMethodException ex) {
									writeError("Object type \""
											+ currentClass.getName()
											+ "\" does not support method \""
											+ methodName
											+ "\" with "
											+ paraClassArray.length
											+ " parameters.");
								}
								catch (InvocationTargetException ex) {
									writeError("Unable to invoke method: " + ex);
								}
							}
							argumentVector.removeAllElements(); //always flush - in case of errors!
							if (tok.sval.equalsIgnoreCase("end")) {
								if (currentInstance == scene) {
									scene.prepare();
									return scene;
								}

								if (!ommitCurrentObject)
									scene.store(
										((NamedObject)currentInstance).getName(),
										(NamedObject)currentInstance);

								textIsNewSurface = true;
								ommitCurrentObject = false;
							}
							methodName = getMethodName(tok.sval);
						}
						break;

					case StreamTokenizer.TT_NUMBER :
						argumentVector.add(new Float(tok.nval));
						break;
				}
				tok.pushBack();
			}
			while (tok.nextToken() != StreamTokenizer.TT_EOF);
			return scene;
		}
		catch (IOException ex) {
			writeError("Error reading scene file: " + ex);
		}
		catch (IllegalAccessException ex) {
			writeError("Surface not appropriate: " + ex);
		}
		return null;
	}

    private static void writeError(String error) {
        ControlFrame.addConsoleText("ERROR: "+error, Constants.failColor);
        System.err.println(error);
    }

    private static String getMethodName(String string) {
		String sval = string;
		if (sval.endsWith("-")) {
			sval = sval.replace('-', ' ').trim();
		}
		return "set"
			+ sval.substring(0, 1).toUpperCase()
			+ sval.substring(1);
	}
}
