package raytracer.internal;

/**
 *
 * Represents a color with values for red, green, blue ranging from 0 to 1
 *
 */
public class Color {

	public static final Color white = new Color(1, 1, 1);
	public float r;
	public float g;
	public float b;

	/**
	 * Creates a color from the integer rgb value
	 * @param rgb integer value
	 */
	public Color(int rgb) {
		this.r = (((rgb <<  8) >> 24) & 0xFF)/255f;
		this.g = (((rgb << 16) >> 24) & 0xFF)/255f;
		this.b = (((rgb << 24) >> 24) & 0xFF)/255f;
	}

	/**
	 * Creates a color object with the given intensities
	 * @param r intensity for red
	 * @param g intensity for green
	 * @param b intensity for blue
	 */
	public Color(float r, float g, float b) {
		this.r = r;
		this.g = g;
		this.b = b;
	}

	/**
	 * Copy constructor
	 * @param col color object to duplicate
	 */
	public Color(Color col) {
		this.r = col.r;
		this.g = col.g;
		this.b = col.b;
	}

	/**
	 * Adds a color
	 * @param col Color to add
	 */
	public void add(Color col) {
		this.r += col.r;
		this.g += col.g;
		this.b += col.b;

	}

	/**
	 * Adds a color with the given intensity
	 * @param obj Color to add
	 * @param intensity Intensity factor applied to all components of the added color
	 */
	public void add(Color obj, float intensity) {
		this.r += obj.r * intensity;
		this.g += obj.g * intensity;
		this.b += obj.b * intensity;
	}

	/**
	 *
	 * Adds a color with the given intensity
	 * @param obj Color to add
	 * @param intensity Intensity color applied componentwise of the added color
	 */
	public void add(Color obj, Color intensity) {
		this.r += obj.r * intensity.r;
		this.g += obj.g * intensity.g;
		this.b += obj.b * intensity.b;
	}

	public void add(Color obj, Color light, float intensity) {
		this.r += obj.r * intensity * light.r;
		this.g += obj.g * intensity * light.r;
		this.b += obj.b * intensity * light.r;
	}

	// @todo ok?
	public void transmit(Color obj, Color transmission) {
		this.r *= obj.r * transmission.r;
		this.g *= obj.g * transmission.g;
		this.b *= obj.b * transmission.b;
	}

	public void mul(Color c) {
		this.r *= c.r;
		this.g *= c.g;
		this.b *= c.b;
	}

	public void mul(Color c, float intensity) {
		this.r *= c.r * intensity;
		this.g *= c.g * intensity;
		this.b *= c.b * intensity;
	}

	/**
	 * Calculates how much the color would influence an existing color (usually the distance to zero)
	 * @return Influence factor
	 */
	public float influence() {
		return (float) Math.sqrt(r * r + g * g + b * b);
	}

	public float getR() {
		return r;
	}

	public float getG() {
		return g;
	}

	public float getB() {
		return b;
	}

	public void normalize() {
		r = Math.max(0, (Math.min(1, r)));
		g = Math.max(0, (Math.min(1, g)));
		b = Math.max(0, (Math.min(1, b)));
	}

	/**
	 * @return the color as RGB integer
	 */
	public int getRGBAsInt() {
		return ((int) (r * 255) << 16)
			+ ((int) (g * 255) << 8)
			+ (int) (b * 255);
	}

	public void print() {
		System.out.println(
			"ints "
				+ (int) (r * 255)
				+ " "
				+ (int) (g * 255)
				+ " "
				+ (int) (b * 255)
				+ " floats: "
				+ r
				+ " "
				+ g
				+ " "
				+ b);
	}

	/**
	 * @param color
	 * @return true if color values are the same
	 */
	public boolean same(Color color) {
		return (Math.abs(color.r - r)*255f < 2)
			&& (Math.abs(color.g - g)*255f < 2)
			&& (Math.abs(color.b - b)*255f < 2);
	}
}
