package raytracer.internal;

/**
 *
 * A 4x4 matrix for operations with homogenous coordinates in three dimensions
 *
 */
public class Matrix {
    private static final int n = 4;
    public float[][] matrix = new float[n][n];

    /**
     * Creates an empty (not unified!) matrix
     */
    public Matrix() {
    }

    /**
     * Creates a matrix with the given paramters. a-d are the elements in the first row, e-h second row...
     */
    public Matrix(
            float a, float b, float c, float d,
            float e, float f, float g, float h,
            float i, float j, float k, float l,
            float m, float n, float o, float p) {

        matrix[0][0] = a;
        matrix[1][0] = b;
        matrix[2][0] = c;
        matrix[3][0] = d;

        matrix[0][1] = e;
        matrix[1][1] = f;
        matrix[2][1] = g;
        matrix[3][1] = h;

        matrix[0][2] = i;
        matrix[1][2] = j;
        matrix[2][2] = k;
        matrix[3][2] = l;

        matrix[0][3] = m;
        matrix[1][3] = n;
        matrix[2][3] = o;
        matrix[3][3] = p;
    }

    /**
     * Returns an entry from the matrix
     * @param row Row of the requested entry
     * @param col Column of the requested entry
     * @return Matrix entry
     */
    public float get(int row, int col) {
        return matrix[col][row];
    }

    /**
     * Set an entry to the matrix
     * @param row Row of the set entry
     * @param col Column of the set entry
     * @param val the new Value
     */
    public void set(int row, int col, float val) {
        matrix[col][row] = val;
    }

    /**
     * Initializes the matrix as a unit matrix
     */
    public void unify() {
        matrix[0][0] = 1;
        matrix[1][0] = 0;
        matrix[2][0] = 0;
        matrix[3][0] = 0;

        matrix[0][1] = 0;
        matrix[1][1] = 1;
        matrix[2][1] = 0;
        matrix[3][1] = 0;

        matrix[0][2] = 0;
        matrix[1][2] = 0;
        matrix[2][2] = 1;
        matrix[3][2] = 0;

        matrix[0][3] = 0;
        matrix[1][3] = 0;
        matrix[2][3] = 0;
        matrix[3][3] = 1;
    }

    /**
     * @return a new matrix object which is A*THIS
     */
    public Matrix mulLeft(Matrix a) { //A*THIS
        return mul(a, this);
    }

    /**
     * @return a new matrix object which is THIS*A
     */
    public Matrix mulRight(Matrix a) { //THIS*A
        return mul(this, a);
    }

    /**
     * @return a new matrix object which is A<sup>getT</sup>*THIS
     */
    public Matrix mulLeftTrans(Matrix a) { //A^getT*THIS
        return mul(a.transpose(), this);
    }

    /**
     * @return a new matrix object which is THIS*A<sup>getT</sup>
     */
    public Matrix mulRightTrans(Matrix a) { //THIS*A^getT
        return mul(this, a.transpose());
    }

    /**
     * Matrix multiplication cMat = aMat * bMat
     * @param aMat
     * @param bMat
     * @return aMat * bMat
     */
    private static final Matrix mul(Matrix aMat, Matrix bMat) {
        Matrix res = new Matrix();
        float c[][] = res.matrix;
        float b[][] = aMat.matrix;
        float a[][] = bMat.matrix;

        for (int i = 0; i < n; i++) {
            for (int j = 0; j < n; j++) {
                c[i][j] = 0;
                for (int k = 0; k < n; k++) {
                    c[i][j] += a[i][k] * b[k][j];
                }
            }
        }
        return res;
    }

    /**
     * @return THIS^getT
     */
    public final Matrix transpose() {
        Matrix res = new Matrix();
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < n; j++) {
                res.matrix[i][j] = matrix[j][i];
            }
        }
        return res;
    }

    /**
     * @return inverse of THIS
     */
    public Matrix inverse() {
        float det = det();
        if (Math.abs(det) < Constants.OUTSIDE_EPSILON) {
            System.out.println("Matrix is singular and has no inverse!");
            return null;
        }
        Matrix res = new Matrix();

        for (int i = 0; i < n; i++) {
            for (int j = 0; j < n; j++) {
                res.matrix[i][j] = (float) (Math.pow(-1, i+j)*coef(i, j) / det);
            }
        }
        return res;
    }

    /**
     * @return det(THIS)
     */
    public float det() {
        return detValue(matrix);
    }


    /**
     * @param row
     * @param col
     * @return det(THIS) ignoring <b>row</b> and <b>col</b>
     */
    private float coef(int row, int col) {
        float a[][] = new float[n-1][n-1];
        int x = 0;
        int y = 0;
        for (int i = 0; i < n; i++)
            if (i != col) {
                y = 0;
                for (int j = 0; j < n; j++)
                    if (j != row) {
                        a[x][y] = matrix[i][j];
                        y++;
                    }
                x++;
            }
        return detValue(a);
    }

    /** This method returns the value of the determinant, using recursion
     @param a the determinant as a nxn float array
     @return the value of the determinant
     */
    public static float detValue(float[][]a)
    {
        int n=a[0].length;
        if (n==1)
        {
            return a[0][0];
        }
        if (n==2)
        {
            return ((a[0][0]*a[1][1])-(a[1][0]*a[0][1]));
        }
        else
        {
            float s=0;
            float[][]temp=new float[n-1][n-1];
            int tempi=0;
            int tempj=0;
            for (int i=0;i<n;i++)
            {
                for (int j=1;j<n;j++)
                {
                    for (int k=0;k<n;k++)
                    {
                        if (k!=i)
                        {
                            temp[tempi][tempj]=a[j][k];
                            tempj++;
                        }
                    }
                    tempi++;
                    tempj=0;
                }
                if(i%2==0)
                    s+=a[0][i]*detValue(temp);
                else
                    s-=a[0][i]*detValue(temp);
                tempi=0;
                tempj=0;
            }
            return s;
        }
    }

    public void print() {
        System.out.println("M =");
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < n; j++) {
                System.out.print("\t"+matrix[j][i]);
            }
            System.out.println();
        }
    }

    public boolean same(Matrix a) {
        for (int i = 0; i < n; i++)
            for (int j = 0; j < n; j++)
                if (!Utils.sameFloat(a.matrix[i][j], matrix[i][j]))
                    return false;
        return true;
    }

    /**
     * Rotates counter-clockwise along the axis given by (0,0,0) and (x,y,z)
     * @param x
     * @param y
     * @param z
     * @param angle
     * @return
     */
    public static Matrix getRotationMatrix(float x, float y, float z, float angle) {
        Matrix m = new Matrix();
        m.unify();

        float c = (float)Math.cos(angle);
        float s = (float)Math.sin(angle);
        float t = 1-c;
        final float txz = t * x * z;
        final float tyz = t * y * z;
        final float txy = t * x * y;

        m.set(0, 0, t * x * x + c);
        m.set(0, 1, txy - s * z);
        m.set(0, 2, txz + s * y);

        m.set(1, 0, txy + s *z);
        m.set(1, 1, t * y * y + c);
        m.set(1, 2, tyz - s * x);

        m.set(2, 0, txz - s * y );
        m.set(2, 1, tyz + s * x );
        m.set(2, 2, t * z * z + c);

        return m;
    }

    public static Matrix getScaleMatrix(float sx, float sy, float sz) {
        Matrix m = new Matrix();
        m.unify();
        m.set(0, 0, sx);
        m.set(1, 1, sy);
        m.set(2, 2, sz);
        return m;
    }

    public static Matrix getTranslationMatrix(float tx, float ty, float tz) {
        Matrix m = new Matrix();
        m.unify();
        m.set(0, 3, tx);
        m.set(1, 3, ty);
        m.set(2, 3, tz);
        return m;
    }
}
