package raytracer.internal;

/**
 *
 * A vector in three dimensions
 *
 */
public class Vector extends Tuple3f {
    /**
     * Creates the vector (x,y,z)
     */
    public Vector(float x, float y, float z) {
        super(x, y, z);
    }

    /**
     * Creates the vector v(x,y,z)
     */
    public Vector(Tuple3f v) {
        super(v);
    }

    /**
     * Calculates the length of the vector
     * @return Length
     */
    public float length() {
        return (float) Math.sqrt(x * x + y * y + z * z);
    }

    /**
     * Calculates the length of the vector
     * @return Length^2
     */
    public float length2() {
        return x * x + y * y + z * z;
    }

    /**
     * Scales the vector such that the length same one
     * @return New vector object with the same direction and length one
     */
    public Vector normalize() {
        final float len = this.length();
        return new Vector(this.x / len, this.y / len, this.z / len);
    }

    /**
     * Scales the vector by the given factor
     * @param t Scaling factor
     * @return New vector object (vector * getT)
     */
    public Vector scale(float t) {
        return new Vector(this.x * t, this.y * t, this.z * t);
    }

    /**
     * Calculates the vector dot product
     * @param v Vector for multiplication
     * @return Dot product
     */
    public float dotProduct(Tuple3f v) {
        return this.x * v.x + this.y * v.y + this.z * v.z;
    }

    /**
     * Calculates the vector cross product
     * @param v Vector for multiplication
     * @return Cross product
     */
    public Vector crossProduct(Tuple3f v) {
        return new Vector(y * v.z - z * v.y, z * v.x - x * v.z, x * v.y - y * v.x);
    }

    /**
     * Inverts a vector componentwise
     * @return New inverted vector
     */
    public Vector invert() {
        return new Vector(-x, -y, -z);
    }


    /**
     * Reflects a vector at a plane with the given normal
     * @param normal Normal vector of the plane to reflect at
     * @return New reflected vector
     */
    public Vector reflect(Vector normal) {
        // result = this - 2*normal*(normal .dot. this)
        final float scale = -2 * dotProduct(normal);
        final Vector v = normal.scale(scale);
        return new Vector(x + v.x, y + v.y, z + v.z);
    }

    public Vector scaleAdd(float s, Tuple3f v) {
        return new Vector(
                this.x + s*v.x,
                this.y + s*v.y,
                this.z + s*v.z);
    }
}
