package raytracer.objects;

import raytracer.internal.*;

/**
 * Created by IntelliJ IDEA.
 * User: Bogdy
 * Date: 18.05.2003
 * Time: 11:55:55
 * To change this template use Options | File Templates.
 */
public class Plane extends Surface {
    private Vector normal = new Vector(0, 1, 0);
    private Point center = new Point(0, 0, 0);
    private Vector texVecU = new Vector(1, 0, 0);
    private Vector texVecV = new Vector(0, 0, 1);
    private float d = 0;

    private void recomputeD() {
        d = -normal.dotProduct(center);

        // project the point on the plane (if point not on the plane)
        Point p = center.add(texVecU);
        float t = p.sub(center).dotProduct(normal);
        if (isInside(p))
            t = -t;

        p.x += t*normal.x;
        p.y += t*normal.y;
        p.z += t*normal.z;

        texVecU = p.sub(center).normalize();
        texVecV = texVecU.crossProduct(normal).normalize();
    }

    /**
     * Plane equation nx(x-cx) + ny(y-cy) + nz(z-cz) = 0
     * @param nx
     * @param ny
     * @param nz
     */
    public void setNormal(float nx, float ny, float nz) {
        normal.x = nx;
        normal.y = ny;
        normal.z = nz;
        normal = normal.normalize();

        recomputeD();
//        normal.print();
    }

    /**
     * Plane equation nx(x-cx) + ny(y-cy) + nz(z-cz) = 0
     * Texture mapping point (0,0)
     * @param cx
     * @param cy
     * @param cz
     */
    public void setCenter(float cx, float cy, float cz) {
        center.x = cx;
        center.y = cy;
        center.z = cz;

        recomputeD();
//        center.print();
    }

    /**
     * Texture mapping point (1,1)
     * @param px
     * @param py
     * @param pz
     */
    public void setTextureVectorU(float px, float py, float pz) {
        texVecU.x = px;
        texVecU.y = py;
        texVecU.z = pz;

        recomputeD();
    }

    /**
     * Calculate the first intersection of a given ray with this object
     * @param ray Ray to calculate intersection with
     * @return Distance between ray starting point and intersection point along the ray direction,
     * INFINITY if no intersection occours.
     * Negative lengths are ignored.
     */
    public Hit intersectRay(Ray ray) {
        return Utils.firstPositiveValue(intersect(ray));
    }

    /**
     * Calculate all intersections of a given ray with this object
     * @param ray Ray to calculate intersection with
     * @return Array of distances between ray starting point and intersection point
     * along the ray direction, null if no intersection occours.
     * Negative lengths included.
     */
    public Hit[] intersect(Ray ray) {
        final Hit[] hits = new Hit[2];
        final float cosAlpha = ray.direction.dotProduct(normal);

        float t = -1;
        // parallel to the plane (alpha=90)
        if (!Utils.sameFloat(cosAlpha, 0))
            t = (d - normal.dotProduct(ray.origin))/cosAlpha;

        if (isInside(ray.pointAt(10000))) {
            hits[0] = new Hit(ray, this, t, true, material);
            hits[1] = new Hit(ray, this, Constants.INFINITY, false, material);
        } else {
            hits[0] = new Hit(ray, this, -Constants.INFINITY, false, material);
            hits[1] = new Hit(ray, this, t, true, material);
        }
        return hits;
    }

    /**
     * Determines whether a given point is inside this object or not
     * @param p Query point
     * @return true, if origin is inside this object, otherwise false
     */
    public boolean isInside(Point p) {
        return normal.dotProduct(p) + d < Constants.OUTSIDE_EPSILON;
    }

    /**
     * Calculates the surface normal at the given point
     * @param hit Point to calculate the normal at
     * @return Normal vector in point origin
     */
    public Vector calcNormal(Hit hit) {
        return normal;
    }

    /**
     * Applies a given transformation to this object
     * @param t Transformation (matrix)
     * @return Should always return "this"
     */
    public Surface apply(Transformation t) {
        normal = new Vector(normal.transform(t));
        center = new Point(center.transform(t));
        recomputeD();
        return this;
    }

    /**
     * @param p
     * @return texture coordinates mapped to the point p
     */
    public TexturePoint calcTextureParam(Point p) {
        final Vector posVec = p.sub(center);
        return new TexturePoint(posVec.dotProduct(texVecU), posVec.dotProduct(texVecV));
    }

    public NamedObject setAttributes(NamedObject newObject) {
        super.setAttributes(newObject);

        final Plane plane = (Plane)newObject;
        plane.setNormal(normal.x, normal.y, normal.z);
        plane.setCenter(center.x, center.y, center.z);
        plane.setTextureVectorU(texVecU.x, texVecU.y, texVecU.z);
        return newObject;
    }

    public NamedObject duplicate() {
        return setAttributes(new Plane());
    }
}
