package raytracer.objects;

import raytracer.internal.*;

/**
 * Created by IntelliJ IDEA.
 * User: Bogdy
 * Date: 30.04.2003
 * Time: 16:13:51
 * To change this template use Options | File Templates.
 */
public class Quadric extends Surface {
    private float A, B, C, D, E, F, G, H, I, J;
    private Matrix matrix = null;
    private double max = Double.MIN_VALUE;
    private Sphere mapSphere = null;

    public Quadric() {
        mapSphere = new Sphere();
        setMatrix(new Matrix());
    }

    public Matrix getMatrix() {
        return new Matrix(
                A, D/2f, E/2f, G/2f,
                D/2f, B, F/2f, H/2f,
                E/2f, F/2f, C, I/2f,
                G/2f, H/2f, I/2f, J);
    }

    private void setMatrix(Matrix matrix) {
        final float[][] m = matrix.matrix;
        setParameter(
                // A        B        C        D
                m[0][0], m[1][1], m[2][2], m[1][0],
                // E        F        G        H
                m[2][0], m[2][1], m[3][0], m[3][1],
                // I        J
                m[3][2], m[3][3]);
    }

    /**
     * Sets the quadric parameters.
     * F(x, y, z) = Ax2 + By2 + Cz2 + 2Dxy + 2Exz + 2Fyz + 2Gx + 2Hy + 2Iz + J = 0
     * INTERNALLY, D-I will be saved divided by 2; so internal equation is:
     * F(x, y, z) = Ax2 + By2 + Cz2 + Dxy + Exz + Fyz + Gx + Hy + Iz + J = 0
     * @param a-j new quadric matrix
     */
    public void setParameter(float a, float b, float c, float d, float e, float f, float g, float h, float i, float j) {
        A = a; B = b; C = c;
        D = d*2f; E = e*2f; F = f*2f;
        G = g*2f; H = h*2f; I = i*2f;
        J = j;
        matrix = new Matrix(
                A, D, E, G,
                D, B, F, H,
                E, F, C, I,
                G, H, I, J);
    }

    /**
     * Calculate the first intersection of a given ray with this object
     * @param ray Ray to calculate intersection with
     * @return Distance between ray starting point and intersection point along the ray direction,
     * INFINITY if no intersection occours
     */
    public Hit intersectRay(Ray ray) {
        return Utils.firstPositiveValue(intersect(ray));
    }

    /**
     * Calculate all intersections of a given ray with this object
     * @param ray Ray to calculate intersection with
     * @return Array of distances between ray starting point and intersection point
     * along the ray direction, null if no intersection occours.
     * Negative lengths included.
     */
    public Hit[] intersect(Ray ray) {
        final Vector d = ray.direction;
        final Point o = ray.origin;

        final float Aq = A*d.x*d.x + B*d.y*d.y + C*d.z*d.z + D*d.x*d.y + E*d.x*d.z + F*d.y*d.z;

        final float Bq = 2f*A*o.x*d.x + 2f*B*o.y*d.y + 2f*C*o.z*d.z + D*(o.x*d.y + o.y*d.x) +
                E*o.x*d.z + F*(o.y*d.z + d.y*o.z) + G*d.x + H*d.y + I*d.z;

        final float Cq = A*o.x*o.x + B*o.y*o.y + C*o.z*o.z + D*o.x*o.y + E*o.x*o.z +
                F*o.y*o.z + G*o.x + H*o.y + I*o.z + J;

        // the solutions:
        if (Utils.sameFloat(Aq, 0)) {
            final Hit[] hits = new Hit[2];
            hits[0] = new Hit(ray, this, -Cq / Bq, true, material);
            hits[1] = new Hit(ray, this, -Cq / Bq, false, material);
            return hits; // one intersection
        }

        final float delta2 = Bq*Bq - 4f*Aq*Cq;
        if (delta2 < 0)
            return null; // no intersection

        final float sgn = Utils.sgn(Bq);
        final float q = -.5f * (float)(Bq + sgn * Math.sqrt(delta2));

        // two intersections
        final Hit[] hits = new Hit[2];
        float t0 = q / Aq;
        float t1 = Cq / q;
        if (t0 > t1) {
            t0 = Cq / q;
            t1 = q / Aq;
        }
        hits[0] = new Hit(ray, this, t0, true, material);
        hits[1] = new Hit(ray, this, t1, false, material);
        if (hits[1].getT() < Constants.OUTSIDE_EPSILON)
            return null; // no intersection
        return hits;
    }

    /**
     * Determines whether a given point is inside this object or not
     * @param p Query point
     * @return true, if origin is inside this object, otherwise false
     */
    public boolean isInside(Point p) {
        // whithout the factor 2 for D-I (see setParam)
        return A*p.x*p.x + B*p.y*p.y + C*p.z*p.z + D*p.x*p.y +
                E*p.x*p.z + F*p.y*p.z + G*p.x + H*p.y + I*p.z +
                J < Constants.OUTSIDE_EPSILON;
    }

    /**
     * Debug
     */
    public void tst() {
        System.out.println("max="+max);
    }

    /**
     * Calculates the surface normal at the given point
     * @param hit Point to calculate the normal at
     * @return Normal vector in point origin
     */
    public Vector calcNormal(Hit hit) {
        Point p = hit.getPoint();
        // whithout the factor 2 for D-I (see setParam)
        final float xn = 2f*A*p.x + D*p.y + E*p.z + G;
        final float yn = 2f*D*p.x + B*p.y + F*p.z + H;
        final float zn = 2f*E*p.x + F*p.y + C*p.z + I;
        return new Vector(xn, yn, zn).normalize();
    }

    /**
     * Applies a given transformation to this object
     * @param t Transformation (matrix)
     * @return Should always return "this"
     */
    public Surface apply(Transformation t) {
        final Matrix inverse = t.getInverse();
        setMatrix(inverse.transpose().mulRight(getMatrix()).mulRight(inverse));
        mapSphere.apply(t);
        return this;
    }

    /* (non-Javadoc)
    * @see raytracer.objects.Surface#calcTextureParam(raytracer.internal.Point)
    */
    public TexturePoint calcTextureParam(Point p) {
        //System.out.println("r "+mapSphere.getCenter().sub(p).length());
        return mapSphere.calcTextureParam(p);
    }

    public NamedObject setAttributes(NamedObject newObject) {
        super.setAttributes(newObject);

        final Quadric quadric = (Quadric)newObject;
        quadric.setMatrix(getMatrix());
        quadric.mapSphere = (Sphere)mapSphere.duplicate();
        return newObject;
    }

    public NamedObject duplicate() {
        return setAttributes(new Quadric());
    }

    public boolean same(Quadric quadric) {
        return (matrix.same(quadric.matrix)) &&
                Utils.sameFloat(A, quadric.A) &&
                Utils.sameFloat(B, quadric.B) &&
                Utils.sameFloat(C, quadric.C) &&
                Utils.sameFloat(D, quadric.D) &&
                Utils.sameFloat(E, quadric.E) &&
                Utils.sameFloat(F, quadric.F) &&
                Utils.sameFloat(G, quadric.G) &&
                Utils.sameFloat(H, quadric.H) &&
                Utils.sameFloat(I, quadric.I) &&
                Utils.sameFloat(J, quadric.J);
    }
}
