package raytracer.objects;

import raytracer.internal.*;

/**
 * Created by IntelliJ IDEA.
 * User: Bogdy
 * Date: 30.04.2003
 * Time: 16:29:52
 * To change this template use Options | File Templates.
 */
public class Sphere extends Surface {
    private float radius;
    private float radius2;
    private Point center;

    public Sphere() {
        setCenter(0, 0, 0);
        setRadius(1);
    }

    public void setCenter(float x, float y, float z) {
        center = new Point(x, y, z);
    }

    public void setRadius(float radius) {
        this.radius = radius;
        this.radius2 = radius * radius;
    }

    public float getRadius() {
        return radius;
    }

    public Point getCenter() {
        return new Point(center);
    }

    /**
     * Calculate the first intersection of a given ray with this object
     * @param ray Ray to calculate intersection with
     * @return Distance between ray starting point and intersection point along the ray direction,
     * INFINITY if no intersection occours
     */
    public Hit intersectRay(Ray ray) {
        Vector v = center.sub(ray.origin);
        float b = v.dotProduct(ray.direction);
        float disc = b*b - v.dotProduct(v) + radius2;
        if (disc <= 0f) // doesn'getT hit
            return Hit.NO_INTERSECTION;
        disc = (float)Math.sqrt(disc);
        float t2 = b + disc; // 2nd root
        if (t2 <= Constants.OUTSIDE_EPSILON) // behing ray origin
            return Hit.NO_INTERSECTION;
        float t1 = b - disc; // 1st root
        // add points to the list
        if (t1 > Constants.OUTSIDE_EPSILON) {
            return new Hit(ray, this, t1, true, material); // enter object
        }
        return new Hit(ray, this, t2, false, material); // tangent to the object
    }

    /**
     * Calculate all intersections of a given ray with this object
     * @param ray Ray to calculate intersection with
     * @return Array of distances between ray starting point and intersection point
     * along the ray direction, null if no intersection occours.
     * Negative lengths included.
     */
    public Hit[] intersect(Ray ray) {
        Vector v = center.sub(ray.origin);
        float b = v.dotProduct(ray.direction);
        float disc = b*b - v.dotProduct(v) + radius2;
        if (disc <= 0f) // doesn't hit
            return null;

        disc = (float)Math.sqrt(disc);
        // add points to the list
        final Hit[] hits = new Hit[2];
        hits[0] = new Hit(ray, this, b - disc, true, material); // 1st root
        hits[1] = new Hit(ray, this, b + disc, true, material); // 2nd root
        return hits;
    }

    /**
     * Determines whether a given point is inside this object or not
     * @param p Query point
     * @return true, if center is inside this object, otherwise false
     */
    public boolean isInside(Point p) {
        return center.sub(p).length2() <= radius2 + Constants.OUTSIDE_EPSILON_POW_2;
    }

    /**
     * Calculates the surface normal at the given point
     * @param hit Point to calculate the normal at
     * @return Normal vector in point center
     */
    public Vector calcNormal(Hit hit) {
        return hit.getPoint().sub(center).normalize();
    }

    /**
     * Applies a given transformation to this object
     * @param t Transformation (matrix)
     * @return Should always return "this"
     */
    public Surface apply(Transformation t) {
        center = new Point(center.transform(t));
        return this;
    }

	/* (non-Javadoc)
	 * @see raytracer.objects.Surface#calcTextureParam(raytracer.internal.Point)
	 */
	public TexturePoint calcTextureParam(Point p) {
		Vector n = p.sub(center).normalize();
		//n.print();
		float theta, phi;

		final float S = (float)Math.sqrt(n.x*n.x + n.y*n.y);

		// theta defined between (0, 2*pi)
		if (Utils.sameFloat(S, 0)) {
			theta = (float) (Utils.sgn(n.y) * Math.PI / 2f);
		} else {
			theta = (float)Math.asin(n.y / S);
		}
		if (n.x < 0) {
			theta = (float)(Math.PI - theta);
		}

		// phi defined between (-pi/2,pi/2)
		phi = (float)Math.acos(n.z);

		//System.out.println("theta="+theta+" phi="+phi);

		// u & v defined between 0 and 1
		final float u = (float)(theta / (2f * Math.PI));
		final float v = (float)(phi/Math.PI);

		return new TexturePoint((u < 0) ? 1f + u : u, v);
	}

    public NamedObject setAttributes(NamedObject newObject) {
        super.setAttributes(newObject);

        final Sphere sphere = (Sphere)newObject;
        sphere.setCenter(center.x,  center.y, center.z);
        sphere.setRadius(radius);
        return newObject;
    }

    public NamedObject duplicate() {
        return setAttributes(new Sphere());
    }
}
