package raytracer.objects;

import raytracer.internal.*;

public abstract class Surface extends NamedObject {

    /**
     * The material every object gets when nothing else is specified.
     */
    protected Material material = Constants.standardMaterial;

    /**
     * Calculate the first intersection of a given ray with this object
     * @param ray Ray to calculate intersection with
     * @return Distance between ray starting point and intersection point along the ray direction,
     * INFINITY if no intersection occours.
     * Negative lengths are ignored.
     */
    public abstract Hit intersectRay(Ray ray);

    /**
     * Calculate all intersections of a given ray with this object
     * @param ray Ray to calculate intersection with
     * @return Array of distances between ray starting point and intersection point
     * along the ray direction, null if no intersection occours.
     * Negative lengths included.
     */
    public abstract Hit[] intersect(Ray ray);

    /**
     * Determines whether a given point is inside this object or not
     * @param p Query point
     * @return true, if origin is inside this object, otherwise false
     */
    public abstract boolean isInside(Point p);

    /**
     * Calculates the surface normal at the given point
     * @param hit Point to calculate the normal at
     * @return Normal vector in point origin
     */
    public abstract Vector calcNormal(Hit hit);

    /**
     * Applies a given transformation to this object
     * @param t Transformation (matrix)
     * @return Should always return "this"
     */
    public abstract Surface apply(Transformation t);

	/**
	 * @param p
	 * @return texture coordinates mapped to the point p 
	 */
	public abstract TexturePoint calcTextureParam(Point p);

    //methods for the dynamic loading interface
    public void setMaterial(Material m) {
        this.material = m;
    }

    public void setTranslation(float tx, float ty, float tz) {
        final Transformation t = new Transformation();
        t.setMatrix(Matrix.getTranslationMatrix(tx, ty, tz));
        apply(t);
    }

    public void setScale(float sx, float sy, float sz) {
        final Transformation t = new Transformation();
        t.setMatrix(Matrix.getScaleMatrix(sx, sy, sz));
        apply(t);
    }

    /**
     * Rotates counter-clockwise along the axis given by (0,0,0) and (x,y,z)
     * @param x
     * @param y
     * @param z
     * @param angle
     */
    public void setRotation( float x, float y, float z, float angle ) {
        final Transformation trans = new Transformation();
        trans.setMatrix(Matrix.getRotationMatrix(x, y, z, angle));
        apply(trans);
    }

    /**
     * @return the material of the surface
     */
    public Material getMaterial() {
        return material;
    }

    public NamedObject setAttributes(NamedObject newObject) {
        super.setAttributes(newObject);

        if (material != null)
            ((Surface)newObject).setMaterial((Material)material.duplicate());
        return newObject;
    }
}
